import asyncio
import logging
import os
from itertools import count
from typing import Any, Dict, cast

import fastapi
from transformers import PreTrainedTokenizerFast

from briton.briton import BritonInteractor, BritonInteractorImpl
from briton.checks import trllm_config_check
from briton.constants import (
    DEFAULT_BRITON_PORT,
    DEFAULT_MAX_FSM_WORKERS,
    MODEL_INPUT_TO_BRITON_FIELD,
    OPENAI_COMPATIBLE_TAG,
    TOOL_CALL_IDS,
    TOOL_CALL_TOKENS,
)
from briton.error_handling import grpc_error_handling
from briton.fsm_cache import FsmCache, add_schema_to_cache
from briton.proto import InferenceRequest
from briton.schema import (
    create_completion,
    create_completion_chunks,
    get_prompt,
    update_raw_model_input,
    validate_model_input,
)
from briton.secrets import get_hf_token_or_none
from briton.trtllm_build_config import TrussTRTLLMBuildConfiguration
from briton.truss_monitor import start_monitor


logger = logging.getLogger(__name__)


class Model:
    def __init__(self, **kwargs):
        self._loaded = False
        self._model = None
        self._config = kwargs["config"]
        self._data_dir = kwargs["data_dir"]
        self._secrets = kwargs["secrets"]
        self._request_id_counter = count(start=1)
        self._briton_process = None
        model_metadata = self._config.get("model_metadata", {})
        tags = model_metadata.get("tags", [])
        self._uses_openai_api = OPENAI_COMPATIBLE_TAG in tags
        trllm_config_check(self._config)
        trtllm_config = self._config.get("trt_llm")
        truss_trtllm_build_config = TrussTRTLLMBuildConfiguration(**trtllm_config.get("build"))
        self._base_model = truss_trtllm_build_config.base_model
        self._tp_count = truss_trtllm_build_config.tensor_parallel_count
        self._tokenizer_repository = truss_trtllm_build_config.checkpoint_repository.repo
        self._kv_cache_free_gpu_mem_fraction = (
            truss_trtllm_build_config.kv_cache_free_gpu_mem_fraction
        )
        self._enable_kv_cache_reuse = (
            truss_trtllm_build_config.plugin_configuration.use_paged_context_fmha
        )
        self._enable_chunked_context = truss_trtllm_build_config.enable_chunked_context
        self._max_num_tokens = truss_trtllm_build_config.max_num_tokens
        self._hf_token = get_hf_token_or_none(self._secrets)
        self._lazy_init_done = False
        self._lazy_init_lock = None
        self._stub = None
        self._max_fsm_workers = DEFAULT_MAX_FSM_WORKERS

        # Allow passing briton_interactor for ease of testing
        self._briton_interactor: BritonInteractor = model_metadata.get(
            "briton_interactor", BritonInteractorImpl()
        )
        self._tool_call_token = TOOL_CALL_TOKENS.get(self._base_model)
        self._tool_call_token_id = TOOL_CALL_IDS.get(self._base_model)

    def load(self):
        if self._loaded:
            return

        # TODO(pankaj) Support loading bundled tokenizer rather than from HF
        self._tokenizer = cast(
            PreTrainedTokenizerFast,
            self._briton_interactor.auto_tokenizer_from_pretrained(
                self._tokenizer_repository, hf_token=self._hf_token
            ),
        )

        # We only support Llama and mistral with Briton, for which this should
        # apply.
        assert isinstance(self._tokenizer, PreTrainedTokenizerFast)

        # These are tokens outside of tokenizer.json. We need to pass these to
        # Briton, to pass to rust tokenizer.
        added_token_decoders = self._tokenizer.added_tokens_decoder
        added_tokens = [token for token in added_token_decoders.values()]

        self._check_supports_tool_calls()

        self._fsm_cache = FsmCache(
            self._briton_interactor.fsm_cache_dir(),
            self._tokenizer,
            self._max_fsm_workers,
            self._tool_call_token_id,
        )

        load_briton = self._briton_interactor.load
        load_briton(
            model_name="briton",
            engine_path=self._data_dir,
            hf_tokenizer=self._tokenizer_repository,
            work_dir=self._data_dir,
            fsm_cache_dir=self._fsm_cache.cache_dir,
            kv_cache_free_gpu_mem_fraction=self._kv_cache_free_gpu_mem_fraction,
            port=DEFAULT_BRITON_PORT,
            added_tokens=added_tokens,
            max_num_tokens=self._max_num_tokens,
            enable_chunked_context=self._enable_chunked_context,
            hf_token=self._hf_token,
            tp_count=self._tp_count,
        )
        self._loaded = True

    async def predict(self, model_input: Dict[str, Any], request: fastapi.Request):
        """
        Run inference

        Note that the async nature of this function is a little tricky. Care is
        needed to make sure this function is a regular async function and not an
        async generator, i.e. there shouldn't be any direct yields in this
        function. This is because we need to support both streaming and
        non-streaming cases in this function. We do this by either returning an
        async-generator for the streaming case, or directly the full text for
        the other case. Returning an async generator for non-streaming case
        interferes with the open ai client proxy.
        """

        # TODO(pankaj) Wire up request cancellation
        async def is_cancelled_fn():
            disconnected = await request.is_disconnected()
            if disconnected:
                logger.info("Request disconnected, cancelling.")
            return disconnected

        if not self._lazy_init_done:
            # While this isn't completely safe, the async lock needs to be
            # created within the same async loop where it will be used. Ideally,
            # the proper solution would involve supporting asynchronous load
            # function, but that is not currently supported in Truss. The risk is
            # that multiple initial requests could end up with different lock
            # instances, making the lock ineffective. In practice, this is
            # highly unlikely. This issue could occur if one request executes
            # the line below and then gets preempted, allowing another request
            # to execute the same line. However, since there is no async
            # operation in the following line, it is very unlikely for the
            # request to be preempted at that point.
            if self._lazy_init_lock is None:
                self._lazy_init_lock = asyncio.Lock()

            async with self._lazy_init_lock:
                self._stub = self._briton_interactor.create_grpc_stub(DEFAULT_BRITON_PORT)
                await start_monitor(self._briton_interactor, logger)
                self._lazy_init_done = True

        validated_input = validate_model_input(
            model_input=model_input, supports_tools=self._tool_call_token is not None
        )
        prompt = get_prompt(validated_input, self._tokenizer)
        model_input.pop("messages", None)
        input_ids = self._tokenizer.encode(prompt, add_special_tokens=False)
        prompt_tokens = len(input_ids)

        request_id = self._calc_request_id()
        briton_request = InferenceRequest(
            request_id=request_id,
            input_text=prompt,
        )
        self._update_request_end_id_pad_id(briton_request, model_input)

        schema = validated_input.output_json_schema
        if schema is not None:
            briton_request.output_schema_hash = await add_schema_to_cache(self._fsm_cache, schema)
            force_tools = validated_input.force_tools
            if force_tools is not None:
                briton_request.force_tools = force_tools

        update_raw_model_input(model_input, validated_input)

        set_briton_request_fields_from_model_input(model_input, briton_request)
        for words in ["bad_words", "stop_words"]:
            if words in model_input:
                for word in model_input[words].split(","):
                    getattr(briton_request, words).append(word)

        resp_iter = self._stub.Infer(briton_request)

        # Get model name for openai compatibility
        model_name = validated_input.model if validated_input.model else ""

        # Default for OpenAI API is to not stream, but old Briton default was to stream.
        stream_predicate = (self._uses_openai_api and validated_input.stream) or (
            not self._uses_openai_api and model_input.get("stream", True)
        )

        with grpc_error_handling():
            if stream_predicate:
                completion_tokens = 0

                # Advance the iterator to get the first chunk, to allow any validation error to be thrown
                async for first_chunk in resp_iter:
                    break

                async def generate_after_first_chunk():
                    nonlocal completion_tokens
                    completion_tokens += len(first_chunk.output_ids)
                    yield first_chunk.output_text
                    async for chunk in resp_iter:
                        completion_tokens += len(chunk.output_ids)
                        yield chunk.output_text

                def completion_tokens_fn():
                    nonlocal completion_tokens
                    return completion_tokens

                async def generate_processed_after_first_chunk():
                    output_text = first_chunk.output_text
                    # Try to remove tool call token from first chunk only
                    if self._tool_call_token is not None:
                        output_text = output_text.removeprefix(self._tool_call_token)
                    eos_token = self._eos_token()
                    if eos_token is not None:
                        output_text = output_text.removesuffix(eos_token)
                    yield output_text

                    async for chunk in resp_iter:
                        output_text = chunk.output_text
                        if eos_token is not None:
                            output_text = output_text.removesuffix(eos_token)
                        yield output_text

                if self._uses_openai_api:
                    return create_completion_chunks(
                        req_id=str(request_id),
                        model=model_name,
                        input_text=generate_after_first_chunk(),
                        eos_token=self._eos_token(),
                        tool_token=self._tool_call_token,
                        prompt_tokens=(
                            prompt_tokens if validated_input.include_stream_usage else None
                        ),
                        completion_tokens_fn=(
                            completion_tokens_fn if validated_input.include_stream_usage else None
                        ),
                    )
                else:
                    return generate_processed_after_first_chunk()
            else:
                full_text, completion_tokens = await _collect_text(resp_iter)
                if self._uses_openai_api:
                    return create_completion(
                        req_id=str(request_id),
                        model=model_name,
                        input_text=full_text,
                        eos_token=self._eos_token(),
                        tool_token=self._tool_call_token,
                        prompt_tokens=prompt_tokens,
                        completion_tokens=completion_tokens,
                    )
                else:
                    eos_token = self._eos_token()
                    if eos_token is not None:
                        full_text = full_text.removesuffix(eos_token)
                    if self._tool_call_token is not None:
                        full_text = full_text.removeprefix(self._tool_call_token)
                    return full_text

    def _calc_request_id(self) -> int:
        """Calculate unique request id.

        Not thread safe, but safe to use in single threaded async context. There
        are no async operations here, so this function is unlikely to be
        preempted in the middle. This is important otherwise we may end up with
        duplicate ids.
        """
        return int(str(os.getpid()) + str(next(self._request_id_counter)))

    def _eos_token(self):
        return getattr(self._tokenizer, "eos_token", None)

    def _eos_token_id(self):
        return getattr(self._tokenizer, "eos_token_id", None)

    def _pad_token_id(self):
        return getattr(self._tokenizer, "pad_token_id", None)

    def _update_request_end_id_pad_id(self, request, model_input):
        end_id = model_input.get("end_id", None) or self._eos_token_id()
        if end_id is not None:
            request.end_id = end_id
        pad_id = model_input.get("pad_id", None) or self._pad_token_id()
        if pad_id is not None:
            request.pad_id = pad_id

    def _check_supports_tool_calls(self):
        if (
            self._tool_call_token is None
            or self._tokenizer.convert_tokens_to_ids(self._tool_call_token)
            == self._tokenizer.unk_token_id
        ):
            self._tool_call_token = None
            self._tool_call_token_id = None


def set_briton_request_fields_from_model_input(model_input, briton_request):
    for model_input_key, briton_field in MODEL_INPUT_TO_BRITON_FIELD.items():
        if model_input_key in model_input:
            model_input_value = model_input[model_input_key]
            setattr(briton_request, briton_field, model_input_value)


async def _collect_text(async_text_iter) -> tuple[str, int]:
    full_text = ""
    completion_tokens = 0
    async for delta in async_text_iter:
        completion_tokens += len(delta.output_ids)
        full_text += delta.output_text
    return full_text, completion_tokens



from typing import List

from tokenizers import AddedToken


def serialize_added_tokens_to_config(added_tokens: List[AddedToken]) -> List[str]:
    """Serialize to pbtxt format."""
    lines = ["added_tokens {"]
    for added_token in added_tokens:
        token_lines = serialize_added_token_to_config(added_token)
        lines.extend([f"  {line}" for line in token_lines])
    lines.append("}")
    return lines


def serialize_added_token_to_config(added_token: AddedToken) -> List[str]:
    """Serialize to pbtxt format."""
    fields = [
        f'content: "{added_token.content}"',
        f"single_word: {added_token.single_word}",
        f"lstrip: {added_token.lstrip}",
        f"rstrip: {added_token.rstrip}",
        f"normalized: {added_token.normalized}",
        f"special: {added_token.special}",
    ]
    return [
        "tokens {",
        *[f"  {field}" for field in fields],
        "}",
    ]



import json
import random
import string
import time
from json import JSONDecodeError
from typing import (
    Any,
    AsyncGenerator,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Union,
    cast,
)

import openai.types.chat.chat_completion as chat_completion
import openai.types.chat.chat_completion_chunk as chat_completion_chunk
from fastapi import HTTPException
from jinja2.exceptions import TemplateError
from openai.types.chat import ChatCompletion, ChatCompletionChunk
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from openai.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageToolCall,
    Function,
)
from openai.types.completion_usage import CompletionUsage
from pydantic import BaseModel, Field, ValidationError, model_validator
from transformers import PreTrainedTokenizerFast


class ModelInput(BaseModel):
    """This class mirrors the `CompletionCreateParamsBase` in the `openai-python` repository.

    However, that class is a TypedDict rather than a pydantic model, so we redefine it here
    to take advantage of pydantic's validation features. In addition, we define helper methods
    to get the formatted prompt, tools to use, and response format to adhere to.

    Unsupported parameters:
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-store
      - OpenAI platform specific
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-metadata
      - OpenAI platform specific
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-frequency_penalty
      - Frequency penalty is not currently passed through to briton
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-logit_bias
      - User provided logit biasing is not implemented
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-logprobs
      - Returning log probabilities is not implemented
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-top_logprobs
      - Returning log probabilities is not implemented
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-n
      - Multiple generation is not implemented
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-service_tier
      - OpenAI platform specific
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-stop
      - Stop words are not implemented
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-user
      - OpenAI platform specific
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-function_call
      - Deprecated
    - https://platform.openai.com/docs/api-reference/chat/create#chat-create-functions
      - Deprecated
    """

    class Tool(BaseModel):
        """An element in the top level `tools` field."""

        class Function(BaseModel):
            name: str
            description: str
            parameters: Dict[str, Any]  # `parameters` holds the json schema
            return_: Optional[Dict[str, Any]] = Field(None, alias="return")

        type: Literal["function"]
        function: Function

        @property
        def json_schema(self) -> Dict[str, Any]:
            return {
                "type": "object",
                "properties": {
                    "name": {"const": self.function.name},
                    "parameters": self.function.parameters,
                },
                "required": ["name", "parameters"],
            }

    class ToolChoice(BaseModel):
        """The top level `tool_choice` field."""

        class FunctionChoice(BaseModel):
            name: str

        type: Literal["function"]
        function: FunctionChoice

    class SchemaResponseFormat(BaseModel):
        """The top level `response_format` field."""

        class JsonSchema(BaseModel):
            """`schema_` holds the actual json schema"""

            schema_: Dict[str, Any] = Field(..., alias="schema")

        type: Literal["json_schema"]
        json_schema: JsonSchema

    class JsonResponseFormat(BaseModel):
        type: Literal["json_object"]

    class TextResponseFormat(BaseModel):
        type: Literal["text"]

    class StreamOptions(BaseModel):
        """The top level `stream_options` field."""

        include_usage: bool

    model: Optional[str] = Field("")

    # TODO: Define `Message` objects to mirror `ChatCompletionMessageParam` for validation
    messages: Optional[List[Dict[str, Any]]] = Field(None)
    prompt_: Optional[str] = Field(None, min_length=1, alias="prompt")

    max_tokens_: Optional[int] = Field(None, alias="max_tokens")
    max_completion_tokens: Optional[int] = Field(None)

    stream: Optional[bool] = Field(False)
    stream_options: Optional[StreamOptions] = Field(None)

    seed_: Optional[int] = Field(None, alias="seed")
    random_seed: Optional[int] = Field(None)

    # TODO(bdubayah) : Wire this up to briton
    frequency_penalty: Optional[float] = Field(0)
    presence_penalty: Optional[float] = Field(0)
    length_penalty: Optional[float] = Field(None)

    temperature: Optional[float] = Field(1)
    top_p_: Optional[float] = Field(1, alias="top_p")
    runtime_top_p: Optional[float] = Field(None)
    top_k_: Optional[int] = Field(None, alias="top_k")
    runtime_top_k: Optional[int] = Field(None)

    response_format: Optional[
        Union[SchemaResponseFormat, JsonResponseFormat, TextResponseFormat]
    ] = Field(None)
    tools_: Optional[List[Tool]] = Field(None, alias="tools")
    tool_choice: Optional[Union[Literal["none", "required", "auto"], ToolChoice]] = Field(None)
    parallel_tool_calls: Optional[bool] = Field(True)

    beam_width: Optional[Literal[1]] = Field(None)

    end_id: Optional[int] = Field(None)
    pad_id: Optional[int] = Field(None)

    _generated_seed: Optional[int] = None

    @model_validator(mode="after")
    def top_p_valid(cls, values):
        if values.top_p < 0 or values.top_p > 1:
            raise ValueError("`top_p` must be between 0 and 1.")
        return values

    @model_validator(mode="after")
    def top_k_valid(cls, values):
        if values.top_k is not None and values.top_k < 0:
            raise ValueError("`top_k` must be greater than or equal to 0.")
        return values

    @model_validator(mode="after")
    def messages_not_empty(cls, values):
        messages = values.messages
        if messages is not None and len(messages) == 0:
            raise ValueError("`messages` cannot be empty.")
        return values

    @model_validator(mode="after")
    def oneof_messages_and_prompt_set(cls, values):
        prompt = values.prompt_
        messages = values.messages
        if prompt is None and messages is None:
            raise ValueError("One of `prompt` or `messages` must be specified.")
        if prompt is not None and messages is not None:
            raise ValueError("Only one of `prompt` and `messages` can be specified.")
        return values

    @model_validator(mode="after")
    def max_tokens_and_max_completion_tokens_not_set(cls, values):
        max_tokens = values.max_tokens_
        max_completion_tokens = values.max_completion_tokens
        if max_tokens is not None and max_completion_tokens is not None:
            raise ValueError(
                "Only one of `max_tokens` and `max_completion_tokens` can be specified."
            )
        return values

    @model_validator(mode="after")
    def tools_valid(cls, values):
        tools = values.tools_
        tool_choice = values.tool_choice
        if tools is not None and tool_choice is None:
            values.tool_choice = "auto"
        if tools is not None and len(tools) == 0 and tool_choice != "none":
            raise ValueError("`tools` cannot be empty.")
        if isinstance(tool_choice, cls.ToolChoice) and tool_choice.function.name not in [
            tool.function.name for tool in tools
        ]:
            raise ValueError("`tool_choice` not in `tools`.")
        return values

    @model_validator(mode="after")
    def tools_not_used_with_prompt(cls, values):
        prompt = values.prompt_
        tool_choice = values.tool_choice
        if prompt is not None and tool_choice is not None and tool_choice != "none":
            raise ValueError("`tool_choice` cannot be used with `prompt`.")
        return values

    @model_validator(mode="after")
    def tools_not_used_with_response_format(cls, values):
        response_format = values.response_format
        tool_choice = values.tool_choice
        if response_format is not None and tool_choice is not None and tool_choice != "none":
            raise ValueError("`tools` cannot be used with `response_format`.")
        return values

    @model_validator(mode="after")
    def adjust_temperature_for_greedy_decoding(cls, values):
        if values.temperature == 0:
            values.temperature = 0.01
            values.top_k_ = 0
            values.runtime_top_k = 0
            values.top_p_ = 0.0
            values.runtime_top_p = 0.0
        return values

    @property
    def force_tools(self) -> Optional[bool]:
        if self.tool_choice is not None and self.tool_choice != "none":
            return self.tool_choice == "required" or isinstance(self.tool_choice, self.ToolChoice)
        return None

    @property
    def tools(self) -> Optional[List[Tool]]:
        """Returns the tools to use, dependent on tool_choice."""
        if self.tool_choice is not None and self.tool_choice != "none":
            if self.tools_ is None:
                raise ValueError("`tools` must be specified with `tool_choice`.")
            if isinstance(self.tool_choice, self.ToolChoice):
                return [
                    tool
                    for tool in self.tools_
                    if tool.function.name == self.tool_choice.function.name
                ]
            return self.tools_
        return None

    @property
    def _tool_dicts(self) -> Optional[List[Dict[str, Any]]]:
        """Convenience property to get all tools as plain dicts."""
        return (
            [tool.model_dump(by_alias=True) for tool in self.tools]
            if self.tools is not None
            else None
        )

    @property
    def output_json_schema(self) -> Optional[Dict[str, Any]]:
        """Creates the output json schema based on the response format or tools."""
        if isinstance(self.response_format, self.SchemaResponseFormat):
            return self.response_format.json_schema.schema_
        if isinstance(self.response_format, self.TextResponseFormat):
            return None
        if isinstance(self.response_format, self.JsonResponseFormat):
            return {}
        tools = self.tools
        if tools is not None:
            schema = {
                "type": "array",
                "items": {"anyOf": [tool.json_schema for tool in tools]},
                "minItems": 1,
            }
            if self.parallel_tool_calls == False:
                schema["maxItems"] = 1
            return schema
        return None

    @property
    def max_tokens(self) -> Optional[int]:
        """`max_tokens` was deprecated in favor of `max_completion_tokens`"""
        return self.max_tokens_ if self.max_tokens_ is not None else self.max_completion_tokens

    @property
    def top_p(self) -> Optional[float]:
        """`top_p` was previously named `runtime_top_p` in briton"""
        return self.runtime_top_p if self.runtime_top_p is not None else self.top_p_

    @property
    def top_k(self) -> Optional[int]:
        """`top_k` was previously named `runtime_top_k` in briton"""
        return self.runtime_top_k if self.runtime_top_k is not None else self.top_k_

    @property
    def seed(self) -> int:
        """`seed` was previously named `random_seed` in briton"""
        if self.seed_ is not None:
            return self.seed_
        if self.random_seed is not None:
            return self.random_seed
        if self._generated_seed is None:
            self._generated_seed = random.randint(-(2**63), 2**63 - 1)
        return self._generated_seed

    @property
    def include_stream_usage(self) -> bool:
        return self.stream_options is not None and self.stream_options.include_usage

    @property
    def combined_messages(self) -> List[Dict[str, Any]]:
        """Combine consecutive user messages into a single message to avoid chat template errors."""
        if not self.messages:
            return []

        combined_messages = []
        current_user_message = None

        for message in self.messages:
            if message["role"] == "user":
                if current_user_message:
                    # OpenAI combines user messages with a space and newline
                    current_user_message["content"] += f" \n{message['content']}"
                else:
                    current_user_message = message.copy()
            else:
                if current_user_message:
                    combined_messages.append(current_user_message)
                    current_user_message = None
                combined_messages.append(message)

        if current_user_message:
            combined_messages.append(current_user_message)

        return combined_messages

    def prompt(self, tokenizer: PreTrainedTokenizerFast) -> str:
        """Calculate text prompt from model_input.

        Prompt may be supplied in the input as such or as messages. If messages
        are supplied, they are used to generate the prompt using chat template.
        """
        if self.prompt_ is None:
            if self.messages is None:
                raise ValueError("`messages` must be specified.")
            return cast(
                str,
                tokenizer.apply_chat_template(
                    conversation=self.combined_messages,
                    tools=self._tool_dicts,
                    tokenize=False,
                    add_generation_prompt=True,
                ),
            )
        return self.prompt_


class ModelInputWithoutTools(ModelInput):
    """Input class for models that don't support function calling."""

    @model_validator(mode="after")
    def tools_not_used(cls, values):
        if values.tools is not None:
            raise ValueError("This model does not support `tools`.")
        return values


def get_prompt(model_input: ModelInput, tokenizer: PreTrainedTokenizerFast) -> str:
    try:
        return model_input.prompt(tokenizer)
    except (ValueError, TemplateError) as e:
        raise HTTPException(status_code=400, detail=str(e))


def validate_model_input(model_input: Dict[str, Any], supports_tools: bool = True) -> ModelInput:
    try:
        if supports_tools:
            return ModelInput(**model_input)
        else:
            return ModelInputWithoutTools(**model_input)
    except (ValueError, ValidationError) as e:
        raise HTTPException(status_code=400, detail=str(e))


def update_raw_model_input(model_input: Dict[str, Any], validated_input: ModelInput) -> None:
    """Updates the `model_input` dictionary with values from the `validated_input` object.

    This is need in cases where 1) the field names different between the spec and briton
    and 2) where we override values (in the case of `seed` or `temperature`).

    Parameters:
    - model_input (Dict[str, Any]): The raw model input.
    - validated_input (ModelInput): The validated model input.
    """
    if validated_input.max_tokens is not None:
        model_input["max_tokens"] = validated_input.max_tokens
    if validated_input.top_k is not None:
        model_input["runtime_top_k"] = validated_input.top_k
    if validated_input.top_p is not None:
        model_input["runtime_top_p"] = validated_input.top_p
    if validated_input.temperature is not None:
        model_input["temperature"] = validated_input.temperature
    model_input["random_seed"] = validated_input.seed


def _load_content_json(content: str) -> Any:
    """Safely load the content json from the input text."""
    try:
        return json.loads(content)
    except JSONDecodeError:
        raise HTTPException(status_code=400, detail="Tool call was cut off by max_tokens.")


def _generate_tool_call_id():
    """Mistral expects a length 9 alphanumeric id"""
    characters = string.ascii_letters + string.digits
    return "".join(random.choice(characters) for _ in range(9))


def _create_tool_calls(content: str) -> List[ChatCompletionMessageToolCall]:
    content_json = _load_content_json(content)
    tool_calls = []
    for briton_fn in content_json:
        fn = Function(name=briton_fn["name"], arguments=str(briton_fn["parameters"]))
        tool_call = ChatCompletionMessageToolCall(
            id=_generate_tool_call_id(), function=fn, type="function"
        )
        tool_calls.append(tool_call)
    return tool_calls


def create_completion(
    req_id: str,
    model: str,
    input_text: str,
    eos_token: Optional[str] = None,
    tool_token: Optional[str] = None,
    prompt_tokens: Optional[int] = None,
    completion_tokens: Optional[int] = None,
) -> ChatCompletion:
    created = int(time.time())
    finish_reason = "stop" if eos_token and input_text.endswith(eos_token) else "length"
    content = input_text.removesuffix(eos_token) if eos_token else input_text
    tool_calls = None
    if tool_token is not None and content.startswith(tool_token):
        content = content.removeprefix(tool_token)
        tool_calls = _create_tool_calls(content)
        content = None
    message = ChatCompletionMessage(content=content, role="assistant", tool_calls=tool_calls)
    choice = chat_completion.Choice(finish_reason=finish_reason, index=0, message=message)
    usage = None
    if prompt_tokens is not None and completion_tokens is not None:
        usage = CompletionUsage(
            prompt_tokens=prompt_tokens,
            completion_tokens=completion_tokens,
            total_tokens=prompt_tokens + completion_tokens,
        )
    return ChatCompletion(
        id=req_id,
        choices=[choice],
        created=created,
        model=model,
        object="chat.completion",
        usage=usage,
    )


def _make_sse_chunk(chunk: ChatCompletionChunk) -> str:
    return f"data: {chunk.model_dump_json()}\n\n"


# It's the responsibility of the caller to pass the full content
def _create_tool_call_deltas(
    content: str,
) -> List[chat_completion_chunk.ChoiceDeltaToolCall]:
    content_json = _load_content_json(content)
    tool_call_deltas = []
    for i, briton_fn in enumerate(content_json):
        fn = chat_completion_chunk.ChoiceDeltaToolCallFunction(
            name=briton_fn["name"], arguments=str(briton_fn["parameters"])
        )
        tool_call_delta = chat_completion_chunk.ChoiceDeltaToolCall(
            index=i, id=_generate_tool_call_id(), function=fn, type="function"
        )
        tool_call_deltas.append(tool_call_delta)
    return tool_call_deltas


def _create_completion_chunk(
    id: str,
    created: int,
    model: str,
    content: Optional[str] = None,
    role: Optional[Literal["system", "user", "assistant", "tool"]] = None,
    finish_reason: Optional[
        Literal["stop", "length", "tool_calls", "content_filter", "function_call"]
    ] = None,
    tool_calls: Optional[List[chat_completion_chunk.ChoiceDeltaToolCall]] = None,
    usage: Optional[CompletionUsage] = None,
) -> ChatCompletionChunk:
    delta = chat_completion_chunk.ChoiceDelta(content=content, role=role, tool_calls=tool_calls)
    choice = chat_completion_chunk.Choice(index=0, delta=delta, finish_reason=finish_reason)
    return ChatCompletionChunk(
        id=id,
        choices=[choice],
        created=created,
        model=model,
        object="chat.completion.chunk",
        usage=usage,
    )


async def create_completion_chunks(
    req_id: str,
    model: str,
    input_text: AsyncGenerator[str, None],
    eos_token: Optional[str] = None,
    tool_token: Optional[str] = None,
    prompt_tokens: Optional[int] = None,
    completion_tokens_fn: Optional[Callable[[], int]] = None,
) -> AsyncGenerator[str, None]:
    created = int(time.time())
    start_chunk = _create_completion_chunk(
        id=req_id, created=created, model=model, content="", role="assistant"
    )
    is_first_iter = True
    delta = None
    async for delta in input_text:
        if is_first_iter:
            if tool_token is not None and delta.startswith(tool_token):
                break
            is_first_iter = False
            yield _make_sse_chunk(start_chunk)
        content = delta.removesuffix(eos_token) if eos_token else delta
        if len(content) == 0:
            continue  # Don't send empty chunks
        chunk = _create_completion_chunk(id=req_id, created=created, model=model, content=content)
        yield _make_sse_chunk(chunk)

    # Handle function call case
    if is_first_iter and delta is not None and tool_token and delta.startswith(tool_token):
        full_text = delta.removeprefix(tool_token)
        async for delta in input_text:
            full_text += delta
        tool_calls = _create_tool_call_deltas(
            full_text.removesuffix(eos_token) if eos_token else full_text
        )
        chunk = _create_completion_chunk(
            id=req_id, created=created, model=model, tool_calls=tool_calls
        )
        yield _make_sse_chunk(start_chunk)
        yield _make_sse_chunk(chunk)
        finish_reason = "tool_calls"
    else:
        finish_reason = (
            "stop" if delta is not None and eos_token and delta.endswith(eos_token) else "length"
        )
    end_chunk = _create_completion_chunk(
        id=req_id, created=created, model=model, finish_reason=finish_reason
    )
    yield _make_sse_chunk(end_chunk)
    if prompt_tokens is not None and completion_tokens_fn is not None:
        completion_tokens = completion_tokens_fn()
        usage = CompletionUsage(
            prompt_tokens=prompt_tokens,
            completion_tokens=completion_tokens,
            total_tokens=prompt_tokens + completion_tokens,
        )
        usage_chunk = _create_completion_chunk(id=req_id, created=created, model=model, usage=usage)
        yield _make_sse_chunk(usage_chunk)
    yield "data: [DONE]\n\n"



# TODO(pankaj) Pick this from truss constants, once a small lib is extracted out of truss.
OPENAI_COMPATIBLE_TAG = "openai-compatible"
DEFAULT_BRITON_PORT = 50051
DEFAULT_TP_COUNT = 1

# Use a directory that can be picked up by baseten-fs, if enabled
FSM_CACHE_DIR = "/cache/model/fsm_cache"

DEFAULT_MAX_FSM_WORKERS = 10

TOOL_CALL_IDS = {
    "llama": 128010,
    "mistral": 5,
}

TOOL_CALL_TOKENS = {
    "llama": "<|python_tag|>",
    "mistral": "[TOOL_CALLS]",
}


MODEL_INPUT_TO_BRITON_FIELD = {
    "max_tokens": "request_output_len",
    "beam_width": "beam_width",
    "repetition_penalty": "repetition_penalty",
    "presence_penalty": "presence_penalty",
    "temperature": "temperature",
    "length_penalty": "len_penalty",
    "end_id": "end_id",
    "pad_id": "pad_id",
    "runtime_top_k": "runtime_top_k",
    "runtime_top_p": "runtime_top_p",
    "random_seed": "random_seed",
}

# TODO(bdubayah): Don't hardcode this
LOCAL_PREDICT_ENDPOINT = "http://localhost:8080/v1/models/model:predict"



import asyncio
from typing import Awaitable, Callable


def get_all_items(queue: asyncio.Queue):
    """Get all items from the queue without waiting."""
    items = []
    try:
        while True:
            item = queue.get_nowait()
            items.append(item)
    except asyncio.QueueEmpty:
        # Once the queue is empty, get_nowait raises QueueEmpty, and we stop
        pass
    return items


def set_event_loop_if_not_exist():
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:  # 'RuntimeError' will be raised if there is no running loop
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    return loop


def retry_predicate(
    pred: Callable[[], Awaitable[bool]],
    max_retries=5,
    delay_secs=2,
) -> Callable[[], Awaitable[bool]]:
    async def wrapper():
        for attempt in range(max_retries):
            try:
                if await pred():
                    return True
            except Exception:
                pass
            if attempt < max_retries - 1:
                await asyncio.sleep(delay_secs)
        return False

    return wrapper


def predicate_with_timeout(
    pred: Callable[[], Awaitable[bool]],
    timeout_secs: int,
) -> Callable[[], Awaitable[bool]]:
    async def wrapper():
        try:
            return await asyncio.wait_for(pred(), timeout=timeout_secs)
        except Exception:
            return False

    return wrapper



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _DataType:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _DataTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_DataType.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    DT_INVALID: _DataType.ValueType  # 0
    DT_INT4: _DataType.ValueType  # 1
    DT_INT8: _DataType.ValueType  # 2
    DT_UINT8: _DataType.ValueType  # 3
    DT_INT32: _DataType.ValueType  # 4
    DT_INT64: _DataType.ValueType  # 5
    DT_FLOAT16: _DataType.ValueType  # 10
    """kHALF"""
    DT_BFLOAT16: _DataType.ValueType  # 11
    """kBF16"""
    DT_FLOAT32: _DataType.ValueType  # 12
    """kFLOAT"""
    DT_FP8: _DataType.ValueType  # 13
    DT_BOOL: _DataType.ValueType  # 20

class DataType(_DataType, metaclass=_DataTypeEnumTypeWrapper):
    """These correspond to nvinfer1::DataType"""

DT_INVALID: DataType.ValueType  # 0
DT_INT4: DataType.ValueType  # 1
DT_INT8: DataType.ValueType  # 2
DT_UINT8: DataType.ValueType  # 3
DT_INT32: DataType.ValueType  # 4
DT_INT64: DataType.ValueType  # 5
DT_FLOAT16: DataType.ValueType  # 10
"""kHALF"""
DT_BFLOAT16: DataType.ValueType  # 11
"""kBF16"""
DT_FLOAT32: DataType.ValueType  # 12
"""kFLOAT"""
DT_FP8: DataType.ValueType  # 13
DT_BOOL: DataType.ValueType  # 20
global___DataType = DataType

@typing.final
class Tensor(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class Shape(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        DIM_FIELD_NUMBER: builtins.int
        @property
        def dim(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
        def __init__(
            self,
            *,
            dim: collections.abc.Iterable[builtins.int] | None = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["dim", b"dim"]) -> None: ...

    SHAPE_FIELD_NUMBER: builtins.int
    DTYPE_FIELD_NUMBER: builtins.int
    DATA_FIELD_NUMBER: builtins.int
    dtype: global___DataType.ValueType
    data: builtins.bytes
    @property
    def shape(self) -> global___Tensor.Shape: ...
    def __init__(
        self,
        *,
        shape: global___Tensor.Shape | None = ...,
        dtype: global___DataType.ValueType = ...,
        data: builtins.bytes = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["shape", b"shape"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["data", b"data", "dtype", b"dtype", "shape", b"shape"]) -> None: ...

global___Tensor = Tensor

@typing.final
class Batch(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    REQUEST_IDS_FIELD_NUMBER: builtins.int
    @property
    def request_ids(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    def __init__(
        self,
        *,
        request_ids: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["request_ids", b"request_ids"]) -> None: ...

global___Batch = Batch

@typing.final
class InferenceRequest(google.protobuf.message.Message):
    """The request message containing the user's name."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    REQUEST_ID_FIELD_NUMBER: builtins.int
    INPUT_TEXT_FIELD_NUMBER: builtins.int
    INPUT_IDS_FIELD_NUMBER: builtins.int
    REQUEST_OUTPUT_LEN_FIELD_NUMBER: builtins.int
    END_ID_FIELD_NUMBER: builtins.int
    PAD_ID_FIELD_NUMBER: builtins.int
    BEAM_WIDTH_FIELD_NUMBER: builtins.int
    TEMPERATURE_FIELD_NUMBER: builtins.int
    RUNTIME_TOP_K_FIELD_NUMBER: builtins.int
    RUNTIME_TOP_P_FIELD_NUMBER: builtins.int
    LEN_PENALTY_FIELD_NUMBER: builtins.int
    REPETITION_PENALTY_FIELD_NUMBER: builtins.int
    PRESENCE_PENALTY_FIELD_NUMBER: builtins.int
    BAD_WORDS_FIELD_NUMBER: builtins.int
    STOP_WORDS_FIELD_NUMBER: builtins.int
    LORA_TASK_ID_FIELD_NUMBER: builtins.int
    LORA_WEIGHTS_FIELD_NUMBER: builtins.int
    LORA_CONFIG_FIELD_NUMBER: builtins.int
    RANDOM_SEED_FIELD_NUMBER: builtins.int
    DRAFT_INPUT_IDS_FIELD_NUMBER: builtins.int
    OUTPUT_SCHEMA_HASH_FIELD_NUMBER: builtins.int
    FSM_STATE_FIELD_NUMBER: builtins.int
    FORCE_TOOLS_FIELD_NUMBER: builtins.int
    WRITE_OUTPUT_STATES_FIELD_NUMBER: builtins.int
    BATCH_FIELD_NUMBER: builtins.int
    request_id: builtins.int
    input_text: builtins.str
    """One of input_text or input_ids should be provided.
    If input_ids is provided then output_ids are returned.
    """
    request_output_len: builtins.int
    end_id: builtins.int
    pad_id: builtins.int
    beam_width: builtins.int
    """Logit selection related"""
    temperature: builtins.float
    runtime_top_k: builtins.int
    runtime_top_p: builtins.float
    len_penalty: builtins.float
    repetition_penalty: builtins.float
    presence_penalty: builtins.float
    lora_task_id: builtins.int
    random_seed: builtins.int
    output_schema_hash: builtins.str
    """Structured output"""
    fsm_state: builtins.int
    force_tools: builtins.bool
    write_output_states: builtins.bool
    @property
    def input_ids(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """TODO(pankaj) Implement this. Not supported now, input_text
        should be provided.
        """

    @property
    def bad_words(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    @property
    def stop_words(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    @property
    def lora_weights(self) -> global___Tensor: ...
    @property
    def lora_config(self) -> global___Tensor: ...
    @property
    def draft_input_ids(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    @property
    def batch(self) -> global___Batch:
        """If request is part of batch, then try to wait for other requests"""

    def __init__(
        self,
        *,
        request_id: builtins.int = ...,
        input_text: builtins.str | None = ...,
        input_ids: collections.abc.Iterable[builtins.int] | None = ...,
        request_output_len: builtins.int | None = ...,
        end_id: builtins.int | None = ...,
        pad_id: builtins.int | None = ...,
        beam_width: builtins.int | None = ...,
        temperature: builtins.float | None = ...,
        runtime_top_k: builtins.int | None = ...,
        runtime_top_p: builtins.float | None = ...,
        len_penalty: builtins.float | None = ...,
        repetition_penalty: builtins.float | None = ...,
        presence_penalty: builtins.float | None = ...,
        bad_words: collections.abc.Iterable[builtins.str] | None = ...,
        stop_words: collections.abc.Iterable[builtins.str] | None = ...,
        lora_task_id: builtins.int | None = ...,
        lora_weights: global___Tensor | None = ...,
        lora_config: global___Tensor | None = ...,
        random_seed: builtins.int | None = ...,
        draft_input_ids: collections.abc.Iterable[builtins.int] | None = ...,
        output_schema_hash: builtins.str | None = ...,
        fsm_state: builtins.int | None = ...,
        force_tools: builtins.bool | None = ...,
        write_output_states: builtins.bool | None = ...,
        batch: global___Batch | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["_batch", b"_batch", "_beam_width", b"_beam_width", "_end_id", b"_end_id", "_force_tools", b"_force_tools", "_fsm_state", b"_fsm_state", "_input_text", b"_input_text", "_len_penalty", b"_len_penalty", "_lora_config", b"_lora_config", "_lora_task_id", b"_lora_task_id", "_lora_weights", b"_lora_weights", "_output_schema_hash", b"_output_schema_hash", "_pad_id", b"_pad_id", "_presence_penalty", b"_presence_penalty", "_random_seed", b"_random_seed", "_repetition_penalty", b"_repetition_penalty", "_request_output_len", b"_request_output_len", "_runtime_top_k", b"_runtime_top_k", "_runtime_top_p", b"_runtime_top_p", "_temperature", b"_temperature", "_write_output_states", b"_write_output_states", "batch", b"batch", "beam_width", b"beam_width", "end_id", b"end_id", "force_tools", b"force_tools", "fsm_state", b"fsm_state", "input_text", b"input_text", "len_penalty", b"len_penalty", "lora_config", b"lora_config", "lora_task_id", b"lora_task_id", "lora_weights", b"lora_weights", "output_schema_hash", b"output_schema_hash", "pad_id", b"pad_id", "presence_penalty", b"presence_penalty", "random_seed", b"random_seed", "repetition_penalty", b"repetition_penalty", "request_output_len", b"request_output_len", "runtime_top_k", b"runtime_top_k", "runtime_top_p", b"runtime_top_p", "temperature", b"temperature", "write_output_states", b"write_output_states"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["_batch", b"_batch", "_beam_width", b"_beam_width", "_end_id", b"_end_id", "_force_tools", b"_force_tools", "_fsm_state", b"_fsm_state", "_input_text", b"_input_text", "_len_penalty", b"_len_penalty", "_lora_config", b"_lora_config", "_lora_task_id", b"_lora_task_id", "_lora_weights", b"_lora_weights", "_output_schema_hash", b"_output_schema_hash", "_pad_id", b"_pad_id", "_presence_penalty", b"_presence_penalty", "_random_seed", b"_random_seed", "_repetition_penalty", b"_repetition_penalty", "_request_output_len", b"_request_output_len", "_runtime_top_k", b"_runtime_top_k", "_runtime_top_p", b"_runtime_top_p", "_temperature", b"_temperature", "_write_output_states", b"_write_output_states", "bad_words", b"bad_words", "batch", b"batch", "beam_width", b"beam_width", "draft_input_ids", b"draft_input_ids", "end_id", b"end_id", "force_tools", b"force_tools", "fsm_state", b"fsm_state", "input_ids", b"input_ids", "input_text", b"input_text", "len_penalty", b"len_penalty", "lora_config", b"lora_config", "lora_task_id", b"lora_task_id", "lora_weights", b"lora_weights", "output_schema_hash", b"output_schema_hash", "pad_id", b"pad_id", "presence_penalty", b"presence_penalty", "random_seed", b"random_seed", "repetition_penalty", b"repetition_penalty", "request_id", b"request_id", "request_output_len", b"request_output_len", "runtime_top_k", b"runtime_top_k", "runtime_top_p", b"runtime_top_p", "stop_words", b"stop_words", "temperature", b"temperature", "write_output_states", b"write_output_states"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_batch", b"_batch"]) -> typing.Literal["batch"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_beam_width", b"_beam_width"]) -> typing.Literal["beam_width"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_end_id", b"_end_id"]) -> typing.Literal["end_id"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_force_tools", b"_force_tools"]) -> typing.Literal["force_tools"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_fsm_state", b"_fsm_state"]) -> typing.Literal["fsm_state"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_input_text", b"_input_text"]) -> typing.Literal["input_text"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_len_penalty", b"_len_penalty"]) -> typing.Literal["len_penalty"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_lora_config", b"_lora_config"]) -> typing.Literal["lora_config"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_lora_task_id", b"_lora_task_id"]) -> typing.Literal["lora_task_id"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_lora_weights", b"_lora_weights"]) -> typing.Literal["lora_weights"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_output_schema_hash", b"_output_schema_hash"]) -> typing.Literal["output_schema_hash"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_pad_id", b"_pad_id"]) -> typing.Literal["pad_id"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_presence_penalty", b"_presence_penalty"]) -> typing.Literal["presence_penalty"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_random_seed", b"_random_seed"]) -> typing.Literal["random_seed"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_repetition_penalty", b"_repetition_penalty"]) -> typing.Literal["repetition_penalty"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_request_output_len", b"_request_output_len"]) -> typing.Literal["request_output_len"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_runtime_top_k", b"_runtime_top_k"]) -> typing.Literal["runtime_top_k"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_runtime_top_p", b"_runtime_top_p"]) -> typing.Literal["runtime_top_p"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_temperature", b"_temperature"]) -> typing.Literal["temperature"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_write_output_states", b"_write_output_states"]) -> typing.Literal["write_output_states"] | None: ...

global___InferenceRequest = InferenceRequest

@typing.final
class InferenceAnswerPart(google.protobuf.message.Message):
    """The response message containing the greetings."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    REQUEST_ID_FIELD_NUMBER: builtins.int
    OUTPUT_TEXT_FIELD_NUMBER: builtins.int
    OUTPUT_IDS_FIELD_NUMBER: builtins.int
    OUTPUT_STATES_FIELD_NUMBER: builtins.int
    request_id: builtins.int
    output_text: builtins.str
    @property
    def output_ids(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    @property
    def output_states(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    def __init__(
        self,
        *,
        request_id: builtins.int = ...,
        output_text: builtins.str = ...,
        output_ids: collections.abc.Iterable[builtins.int] | None = ...,
        output_states: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["output_ids", b"output_ids", "output_states", b"output_states", "output_text", b"output_text", "request_id", b"request_id"]) -> None: ...

global___InferenceAnswerPart = InferenceAnswerPart

@typing.final
class AddedToken(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CONTENT_FIELD_NUMBER: builtins.int
    SINGLE_WORD_FIELD_NUMBER: builtins.int
    LSTRIP_FIELD_NUMBER: builtins.int
    RSTRIP_FIELD_NUMBER: builtins.int
    NORMALIZED_FIELD_NUMBER: builtins.int
    SPECIAL_FIELD_NUMBER: builtins.int
    content: builtins.str
    single_word: builtins.bool
    lstrip: builtins.bool
    rstrip: builtins.bool
    normalized: builtins.bool
    special: builtins.bool
    def __init__(
        self,
        *,
        content: builtins.str | None = ...,
        single_word: builtins.bool | None = ...,
        lstrip: builtins.bool | None = ...,
        rstrip: builtins.bool | None = ...,
        normalized: builtins.bool | None = ...,
        special: builtins.bool | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["_content", b"_content", "_lstrip", b"_lstrip", "_normalized", b"_normalized", "_rstrip", b"_rstrip", "_single_word", b"_single_word", "_special", b"_special", "content", b"content", "lstrip", b"lstrip", "normalized", b"normalized", "rstrip", b"rstrip", "single_word", b"single_word", "special", b"special"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["_content", b"_content", "_lstrip", b"_lstrip", "_normalized", b"_normalized", "_rstrip", b"_rstrip", "_single_word", b"_single_word", "_special", b"_special", "content", b"content", "lstrip", b"lstrip", "normalized", b"normalized", "rstrip", b"rstrip", "single_word", b"single_word", "special", b"special"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_content", b"_content"]) -> typing.Literal["content"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_lstrip", b"_lstrip"]) -> typing.Literal["lstrip"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_normalized", b"_normalized"]) -> typing.Literal["normalized"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_rstrip", b"_rstrip"]) -> typing.Literal["rstrip"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_single_word", b"_single_word"]) -> typing.Literal["single_word"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_special", b"_special"]) -> typing.Literal["special"] | None: ...

global___AddedToken = AddedToken

@typing.final
class AddedTokens(google.protobuf.message.Message):
    """These mirror the added tokens concept in HF PreTrainedTokenizerFast. Most
    tokens are defined in tokenizer.json and are automatically picked up by the
    rust tokenizer. But those defined outside, as in tokenizer_config.json and
    special_tokens_map.json, need to be added manually. Those additional tokens
    are supplied at startup here.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TOKENS_FIELD_NUMBER: builtins.int
    @property
    def tokens(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___AddedToken]: ...
    def __init__(
        self,
        *,
        tokens: collections.abc.Iterable[global___AddedToken] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["tokens", b"tokens"]) -> None: ...

global___AddedTokens = AddedTokens

@typing.final
class BritonConfig(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _BatchSchedulerPolicy:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _BatchSchedulerPolicyEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[BritonConfig._BatchSchedulerPolicy.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        MAX_UTILIZATION: BritonConfig._BatchSchedulerPolicy.ValueType  # 0
        GUARANTEED_NO_EVICT: BritonConfig._BatchSchedulerPolicy.ValueType  # 1

    class BatchSchedulerPolicy(_BatchSchedulerPolicy, metaclass=_BatchSchedulerPolicyEnumTypeWrapper): ...
    MAX_UTILIZATION: BritonConfig.BatchSchedulerPolicy.ValueType  # 0
    GUARANTEED_NO_EVICT: BritonConfig.BatchSchedulerPolicy.ValueType  # 1

    ENGINE_PATH_FIELD_NUMBER: builtins.int
    HF_TOKENIZER_FIELD_NUMBER: builtins.int
    PORT_FIELD_NUMBER: builtins.int
    BATCH_SCHEDULER_POLICY_FIELD_NUMBER: builtins.int
    ENABLE_TRT_OVERLAP_FIELD_NUMBER: builtins.int
    MAX_TOKENS_IN_PAGED_KV_CACHE_FIELD_NUMBER: builtins.int
    KV_CACHE_FREE_GPU_MEM_FRACTION_FIELD_NUMBER: builtins.int
    MEDUSA_DECODING_MODE_FIELD_NUMBER: builtins.int
    ENABLE_CHUNKED_CONTEXT_FIELD_NUMBER: builtins.int
    ENABLE_KV_CACHE_REUSE_FIELD_NUMBER: builtins.int
    KV_CACHE_HOST_MEMORY_BYTES_FIELD_NUMBER: builtins.int
    LORA_CACHE_MAX_ADAPTER_SIZE_FIELD_NUMBER: builtins.int
    LORA_CACHE_OPTIMAL_ADAPTER_SIZE_FIELD_NUMBER: builtins.int
    LORA_CACHE_GPU_MEMORY_FRACTION_FIELD_NUMBER: builtins.int
    LORA_CACHE_HOST_MEMORY_BYTES_FIELD_NUMBER: builtins.int
    FSM_CACHE_DIR_FIELD_NUMBER: builtins.int
    MAX_BATCH_WAIT_MS_FIELD_NUMBER: builtins.int
    ADDED_TOKENS_FIELD_NUMBER: builtins.int
    MAX_NUM_TOKENS_FIELD_NUMBER: builtins.int
    engine_path: builtins.str
    hf_tokenizer: builtins.str
    port: builtins.int
    batch_scheduler_policy: global___BritonConfig.BatchSchedulerPolicy.ValueType
    enable_trt_overlap: builtins.bool
    max_tokens_in_paged_kv_cache: builtins.int
    kv_cache_free_gpu_mem_fraction: builtins.float
    medusa_decoding_mode: builtins.bool
    enable_chunked_context: builtins.bool
    enable_kv_cache_reuse: builtins.bool
    kv_cache_host_memory_bytes: builtins.int
    lora_cache_max_adapter_size: builtins.int
    lora_cache_optimal_adapter_size: builtins.int
    lora_cache_gpu_memory_fraction: builtins.float
    lora_cache_host_memory_bytes: builtins.int
    fsm_cache_dir: builtins.str
    """Path to cache dir with StatesToTokens"""
    max_batch_wait_ms: builtins.int
    """TODO(pankaj) This field is a placeholder for future use. Currently it is
    not used.

    Wait up to this time for all batch requests to arrive. Normally requests
    should arrive pretty quickly, this is to tackle exceptional cases. So this
    value can be relatively large, say a few ms.
    """
    max_num_tokens: builtins.int
    @property
    def added_tokens(self) -> global___AddedTokens:
        """Added tokens to the tokenizer"""

    def __init__(
        self,
        *,
        engine_path: builtins.str = ...,
        hf_tokenizer: builtins.str = ...,
        port: builtins.int | None = ...,
        batch_scheduler_policy: global___BritonConfig.BatchSchedulerPolicy.ValueType | None = ...,
        enable_trt_overlap: builtins.bool | None = ...,
        max_tokens_in_paged_kv_cache: builtins.int | None = ...,
        kv_cache_free_gpu_mem_fraction: builtins.float | None = ...,
        medusa_decoding_mode: builtins.bool | None = ...,
        enable_chunked_context: builtins.bool | None = ...,
        enable_kv_cache_reuse: builtins.bool | None = ...,
        kv_cache_host_memory_bytes: builtins.int | None = ...,
        lora_cache_max_adapter_size: builtins.int | None = ...,
        lora_cache_optimal_adapter_size: builtins.int | None = ...,
        lora_cache_gpu_memory_fraction: builtins.float | None = ...,
        lora_cache_host_memory_bytes: builtins.int | None = ...,
        fsm_cache_dir: builtins.str | None = ...,
        max_batch_wait_ms: builtins.int | None = ...,
        added_tokens: global___AddedTokens | None = ...,
        max_num_tokens: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["_added_tokens", b"_added_tokens", "_batch_scheduler_policy", b"_batch_scheduler_policy", "_enable_chunked_context", b"_enable_chunked_context", "_enable_kv_cache_reuse", b"_enable_kv_cache_reuse", "_enable_trt_overlap", b"_enable_trt_overlap", "_fsm_cache_dir", b"_fsm_cache_dir", "_kv_cache_free_gpu_mem_fraction", b"_kv_cache_free_gpu_mem_fraction", "_kv_cache_host_memory_bytes", b"_kv_cache_host_memory_bytes", "_lora_cache_gpu_memory_fraction", b"_lora_cache_gpu_memory_fraction", "_lora_cache_host_memory_bytes", b"_lora_cache_host_memory_bytes", "_lora_cache_max_adapter_size", b"_lora_cache_max_adapter_size", "_lora_cache_optimal_adapter_size", b"_lora_cache_optimal_adapter_size", "_max_batch_wait_ms", b"_max_batch_wait_ms", "_max_num_tokens", b"_max_num_tokens", "_max_tokens_in_paged_kv_cache", b"_max_tokens_in_paged_kv_cache", "_medusa_decoding_mode", b"_medusa_decoding_mode", "_port", b"_port", "added_tokens", b"added_tokens", "batch_scheduler_policy", b"batch_scheduler_policy", "enable_chunked_context", b"enable_chunked_context", "enable_kv_cache_reuse", b"enable_kv_cache_reuse", "enable_trt_overlap", b"enable_trt_overlap", "fsm_cache_dir", b"fsm_cache_dir", "kv_cache_free_gpu_mem_fraction", b"kv_cache_free_gpu_mem_fraction", "kv_cache_host_memory_bytes", b"kv_cache_host_memory_bytes", "lora_cache_gpu_memory_fraction", b"lora_cache_gpu_memory_fraction", "lora_cache_host_memory_bytes", b"lora_cache_host_memory_bytes", "lora_cache_max_adapter_size", b"lora_cache_max_adapter_size", "lora_cache_optimal_adapter_size", b"lora_cache_optimal_adapter_size", "max_batch_wait_ms", b"max_batch_wait_ms", "max_num_tokens", b"max_num_tokens", "max_tokens_in_paged_kv_cache", b"max_tokens_in_paged_kv_cache", "medusa_decoding_mode", b"medusa_decoding_mode", "port", b"port"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["_added_tokens", b"_added_tokens", "_batch_scheduler_policy", b"_batch_scheduler_policy", "_enable_chunked_context", b"_enable_chunked_context", "_enable_kv_cache_reuse", b"_enable_kv_cache_reuse", "_enable_trt_overlap", b"_enable_trt_overlap", "_fsm_cache_dir", b"_fsm_cache_dir", "_kv_cache_free_gpu_mem_fraction", b"_kv_cache_free_gpu_mem_fraction", "_kv_cache_host_memory_bytes", b"_kv_cache_host_memory_bytes", "_lora_cache_gpu_memory_fraction", b"_lora_cache_gpu_memory_fraction", "_lora_cache_host_memory_bytes", b"_lora_cache_host_memory_bytes", "_lora_cache_max_adapter_size", b"_lora_cache_max_adapter_size", "_lora_cache_optimal_adapter_size", b"_lora_cache_optimal_adapter_size", "_max_batch_wait_ms", b"_max_batch_wait_ms", "_max_num_tokens", b"_max_num_tokens", "_max_tokens_in_paged_kv_cache", b"_max_tokens_in_paged_kv_cache", "_medusa_decoding_mode", b"_medusa_decoding_mode", "_port", b"_port", "added_tokens", b"added_tokens", "batch_scheduler_policy", b"batch_scheduler_policy", "enable_chunked_context", b"enable_chunked_context", "enable_kv_cache_reuse", b"enable_kv_cache_reuse", "enable_trt_overlap", b"enable_trt_overlap", "engine_path", b"engine_path", "fsm_cache_dir", b"fsm_cache_dir", "hf_tokenizer", b"hf_tokenizer", "kv_cache_free_gpu_mem_fraction", b"kv_cache_free_gpu_mem_fraction", "kv_cache_host_memory_bytes", b"kv_cache_host_memory_bytes", "lora_cache_gpu_memory_fraction", b"lora_cache_gpu_memory_fraction", "lora_cache_host_memory_bytes", b"lora_cache_host_memory_bytes", "lora_cache_max_adapter_size", b"lora_cache_max_adapter_size", "lora_cache_optimal_adapter_size", b"lora_cache_optimal_adapter_size", "max_batch_wait_ms", b"max_batch_wait_ms", "max_num_tokens", b"max_num_tokens", "max_tokens_in_paged_kv_cache", b"max_tokens_in_paged_kv_cache", "medusa_decoding_mode", b"medusa_decoding_mode", "port", b"port"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_added_tokens", b"_added_tokens"]) -> typing.Literal["added_tokens"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_batch_scheduler_policy", b"_batch_scheduler_policy"]) -> typing.Literal["batch_scheduler_policy"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_enable_chunked_context", b"_enable_chunked_context"]) -> typing.Literal["enable_chunked_context"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_enable_kv_cache_reuse", b"_enable_kv_cache_reuse"]) -> typing.Literal["enable_kv_cache_reuse"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_enable_trt_overlap", b"_enable_trt_overlap"]) -> typing.Literal["enable_trt_overlap"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_fsm_cache_dir", b"_fsm_cache_dir"]) -> typing.Literal["fsm_cache_dir"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_kv_cache_free_gpu_mem_fraction", b"_kv_cache_free_gpu_mem_fraction"]) -> typing.Literal["kv_cache_free_gpu_mem_fraction"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_kv_cache_host_memory_bytes", b"_kv_cache_host_memory_bytes"]) -> typing.Literal["kv_cache_host_memory_bytes"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_lora_cache_gpu_memory_fraction", b"_lora_cache_gpu_memory_fraction"]) -> typing.Literal["lora_cache_gpu_memory_fraction"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_lora_cache_host_memory_bytes", b"_lora_cache_host_memory_bytes"]) -> typing.Literal["lora_cache_host_memory_bytes"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_lora_cache_max_adapter_size", b"_lora_cache_max_adapter_size"]) -> typing.Literal["lora_cache_max_adapter_size"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_lora_cache_optimal_adapter_size", b"_lora_cache_optimal_adapter_size"]) -> typing.Literal["lora_cache_optimal_adapter_size"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_max_batch_wait_ms", b"_max_batch_wait_ms"]) -> typing.Literal["max_batch_wait_ms"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_max_num_tokens", b"_max_num_tokens"]) -> typing.Literal["max_num_tokens"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_max_tokens_in_paged_kv_cache", b"_max_tokens_in_paged_kv_cache"]) -> typing.Literal["max_tokens_in_paged_kv_cache"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_medusa_decoding_mode", b"_medusa_decoding_mode"]) -> typing.Literal["medusa_decoding_mode"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing.Literal["_port", b"_port"]) -> typing.Literal["port"] | None: ...

global___BritonConfig = BritonConfig

@typing.final
class TokenToNextState(google.protobuf.message.Message):
    """FSM"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class TokenToNextStateEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.int
        value: builtins.int
        def __init__(
            self,
            *,
            key: builtins.int = ...,
            value: builtins.int = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    TOKEN_TO_NEXT_STATE_FIELD_NUMBER: builtins.int
    @property
    def token_to_next_state(self) -> google.protobuf.internal.containers.ScalarMap[builtins.int, builtins.int]: ...
    def __init__(
        self,
        *,
        token_to_next_state: collections.abc.Mapping[builtins.int, builtins.int] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["token_to_next_state", b"token_to_next_state"]) -> None: ...

global___TokenToNextState = TokenToNextState

@typing.final
class StatesToTokens(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class StatesToTokensEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.int
        @property
        def value(self) -> global___TokenToNextState: ...
        def __init__(
            self,
            *,
            key: builtins.int = ...,
            value: global___TokenToNextState | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    STATES_TO_TOKENS_FIELD_NUMBER: builtins.int
    VOCAB_SIZE_FIELD_NUMBER: builtins.int
    EOS_TOKEN_ID_FIELD_NUMBER: builtins.int
    TOOLS_ID_FIELD_NUMBER: builtins.int
    vocab_size: builtins.int
    eos_token_id: builtins.int
    tools_id: builtins.int
    @property
    def states_to_tokens(self) -> google.protobuf.internal.containers.MessageMap[builtins.int, global___TokenToNextState]: ...
    def __init__(
        self,
        *,
        states_to_tokens: collections.abc.Mapping[builtins.int, global___TokenToNextState] | None = ...,
        vocab_size: builtins.int = ...,
        eos_token_id: builtins.int = ...,
        tools_id: builtins.int | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["_tools_id", b"_tools_id", "tools_id", b"tools_id"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["_tools_id", b"_tools_id", "eos_token_id", b"eos_token_id", "states_to_tokens", b"states_to_tokens", "tools_id", b"tools_id", "vocab_size", b"vocab_size"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["_tools_id", b"_tools_id"]) -> typing.Literal["tools_id"] | None: ...

global___StatesToTokens = StatesToTokens
